#include "PLA.h"

static int PLA_MEMORY_SET_ASIDE = 0;
static int PNNL_MALLOC_INITIALIZED = 0;

extern int PLA_MEMORY_SET_ASIDE_TYPE;

extern PLA_Obj_def *PLA_GLOBAL_OBJ_STRUCT = NULL;
int                *PLA_GLOBAL_OBJ_INDEX = NULL;
extern int         *PLA_GLOBAL_VIEWS_STRUCT = NULL;
extern int         *PLA_GLOBAL_VIEWS_INDEX = NULL;
static int         PLA_NUMBER_OF_STRUCTS = 0;

typedef double Align[4]; /* for alignment to double complex boundary */

void *pnnl_malloc(int);
void  pnnl_free(void*);

union header {
   struct {
      union header *ptr; /* next block if on free list */
      int     size; /* size of this block */
   } s;
   Align x; /* force alignment of blocks */
};

typedef union header Header;
static Header  base;          /* empty list to get started */
static Header *freep = NULL ; /* start of free list */
static Header *base_scratch = NULL;

/* pla_malloc : general-purpose storage allocator */

int PLA_pnnl_malloc_init(void *scratch_space, int size_in_bytes)
{
   int nunits;

    printf("PLA_pnnl_malloc_init called with %5ld\n", size_in_bytes); 

   if(PNNL_MALLOC_INITIALIZED)
   {
      printf("The PLA_pnnl_malloc_init routine can only be called once\n");
      return(1);
   }
   else
   {
      nunits = size_in_bytes - sizeof(Header);
      base_scratch = (Header*)scratch_space;
      pnnl_malloc(nunits);
      PNNL_MALLOC_INITIALIZED = 1;
   }
}

void *pnnl_malloc(int nbytes)
{
   Header *p, *prevp;
   int nunits;

   /* printf("pnnl_malloc(%5ld) called\n", (int)nbytes);  */
   nunits = (nbytes+sizeof(Header)-1)/sizeof(Header) + 1;
   if((prevp = freep) == NULL) { /* no free list yet */
      base.s.ptr = freep = prevp = &base;
      base.s.size = 0;
   }
   for(p = prevp->s.ptr; /* empty stmt */ ; prevp = p, p = p->s.ptr) {
      if(p->s.size >= nunits) { /* big enough */
         if(p->s.size == nunits) /* exactly */
            prevp->s.ptr = p->s.ptr;
         else { /* allocate tail end */
            p->s.size -= nunits;
            p += p->s.size;
            p->s.size = nunits;
         }
         freep = prevp;
     /*printf("Allocated %d UNITS\n", p->s.size); */
         return  (void *)(p + 1);
      }
      if (p == freep) /* wrapped around free list */
         if(!PNNL_MALLOC_INITIALIZED)
         {
/* printf("Trying to allocated the space like morecore would\n"); */
            p = (Header*)base_scratch;
            p->s.size = nunits;
            pnnl_free((void*)(p+1));
            p = freep;
            return p;
         }
         else
         {
            printf("Had to return NULL\n");
            return NULL;
         }
   }
}

void *pnnl_calloc(int count, int size)
{
   void *ret_ptr;
   int   i, loop_top;
   char *start_ptr, zero_char;

    /* printf("pnnl_calloc called\n");   */
   loop_top = count*size;
   ret_ptr = pnnl_malloc(loop_top);
   start_ptr = (char*)ret_ptr;
   loop_top = loop_top/sizeof(char);
   zero_char = '\0';
   for(i = 0; i < loop_top; i++)
   {
      *(start_ptr+i) = zero_char;
   }
   return(ret_ptr); 
}

/* free: put block ap in the free list */
void pnnl_free(void *ap)
{
   Header *bp, *p;

   /* printf("pnnl_free called\n"); */
   bp = (Header *)ap - 1; /* point to block header */
/* printf("Free of size %5d UNITS\n", bp->s.size); */
   for(p = freep; !(bp > p && bp < p->s.ptr); p = p->s.ptr)
      if(p >= p->s.ptr && (bp > p || bp < p->s.ptr))
         break; /* freed block at start or end of arena */
   if(bp + bp->s.size == p->s.ptr) { /* join to upper nbr */
      bp->s.size += p->s.ptr->s.size;
      bp->s.ptr = p->s.ptr->s.ptr;
   }else
      bp->s.ptr = p->s.ptr;
   if(p + p->s.size == bp) { /* join to lower nbr */
      p->s.size += bp->s.size;
      p->s.ptr = bp->s.ptr;
   }else
       p->s.ptr = bp;
   freep = p;
}
/* nbr == neighbor */
int PLA_Scratch_set_aside_size(int number_of_structs, int m, int n, int nb_alg, int *size_required)
{
   int internal_size_calc = 0;

   /* printf("At this point the only space set aside as scratch space is the object headers\n"); */

   internal_size_calc += number_of_structs * sizeof(PLA_Obj_def); /* objects */
   internal_size_calc += number_of_structs * sizeof(int); /* object index */
   internal_size_calc += number_of_structs * sizeof(int); /* views */
   internal_size_calc += number_of_structs * sizeof(int); /* views index */
   *size_required = internal_size_calc;
   return(PLA_SUCCESS);   
}

int PLA_Init_object_scratch_space_buf(int number_of_structs, int m, int n,
                                  int nb_alg, int size_in_bytes, void *buffer)
{
   int error = 0;
   int size_required;
   void *follow_ptr;
   int  *int_ptr;
   PLA_Obj obj_ptr;
   if(PLA_MEMORY_SET_ASIDE)
   {
      printf("PLA_Init_object_scratch_space can only be called once\n"); 
      return(1);
   }
   else
   {
      error += PLA_Scratch_set_aside_size(number_of_structs, m, n, nb_alg, &size_required);
      if(size_in_bytes < size_required)
      {
         printf("PLA_Init_scratch_space was passed too little memory %5d vs %5d\n",
                 size_in_bytes, size_required);
         return(1);
      }
      else
      {
         follow_ptr = (void*)buffer; 
         PLA_GLOBAL_OBJ_STRUCT =  (PLA_Obj)follow_ptr;
         obj_ptr = PLA_GLOBAL_OBJ_STRUCT + number_of_structs;
         int_ptr = (int*)obj_ptr; 
         PLA_GLOBAL_OBJ_INDEX  =   int_ptr;
         int_ptr = PLA_GLOBAL_OBJ_INDEX + number_of_structs;
         PLA_GLOBAL_VIEWS_STRUCT = int_ptr;
         int_ptr = PLA_GLOBAL_OBJ_INDEX + number_of_structs;
         PLA_GLOBAL_VIEWS_INDEX  = int_ptr;
         PLA_NUMBER_OF_STRUCTS = number_of_structs;
      }
   }
   PLA_MEMORY_SET_ASIDE = 1;
   PLA_MEMORY_SET_ASIDE_TYPE = PLA_USER_BUFFER; 
   return(error);
}

int PLA_Init_object_scratch_space(int number_of_structs, int m, int n, int nb_alg)
{
   int error = 0;
   int size_required;
   if(PLA_MEMORY_SET_ASIDE)
   {
      printf("PLA_Init_object_scratch_space can only be called once\n"); 
      return(1);
   }
   else
   {
         PLA_GLOBAL_OBJ_STRUCT =  pla_malloc(number_of_structs*sizeof(PLA_Obj_def));
         PLA_GLOBAL_OBJ_INDEX  =  pla_calloc(number_of_structs, sizeof(int));
         PLA_GLOBAL_VIEWS_STRUCT = pla_malloc(number_of_structs*sizeof(int));
         PLA_GLOBAL_VIEWS_INDEX  = pla_calloc(number_of_structs, sizeof(int));
         PLA_NUMBER_OF_STRUCTS = number_of_structs;
         if (  (NULL == PLA_GLOBAL_OBJ_STRUCT ) ||
               (NULL == PLA_GLOBAL_OBJ_INDEX  ) ||
               (NULL == PLA_GLOBAL_VIEWS_STRUCT ) ||
               (NULL == PLA_GLOBAL_VIEWS_INDEX  ) )
         {
            printf("Could not allocate scratch_space\n");
            return(1);
         }
   }
   PLA_MEMORY_SET_ASIDE = 1;
   PLA_MEMORY_SET_ASIDE_TYPE = PLA_HEAP; /* malloced */
   return(error);
}

int PLA_Init_object_structs(int number_of_structs)
{
   void *memory_ptr = NULL;
   int  bytes_required;
   int error = 0;

   error += PLA_Scratch_set_aside_size(number_of_structs, 0, 0, 0, &bytes_required);
   memory_ptr = pla_malloc(bytes_required);  /* FIX THIS KLUDGE JAG */
   if(NULL == memory_ptr)
   {
      printf("PLA_Init_object_structs failed -- could not get enough memory\n");
      return(1);
   }
   else
   {
      pla_free(memory_ptr); /* FIX THIS KLUDGE JAG */
/* PLA_Init_object_scratch_space_buf(number_of_structs,0,0,0,bytes_required,memory_ptr);*/
      PLA_Init_object_scratch_space(number_of_structs, 0, 0, 0); 
      return(PLA_SUCCESS);
   }
   return(PLA_SUCCESS);
}

/* NOTE : The following functions do NOT return error codes but indices */
/* FURTHER NOTE : The term GET is literal -- you "GET" it, you mark it USED */
int PLA_Get_free_object_struct(void)
{
   int i, index = -1;
   for(i = 0; i < PLA_NUMBER_OF_STRUCTS; i++)
      if(PLA_GLOBAL_OBJ_INDEX[i] == 0)
      {
         index = i;
         PLA_GLOBAL_OBJ_INDEX[index] = 1; /* used */
         break;
      }
   return(index);
}

int PLA_Get_free_views_struct(void)
{
   int i, index = -1;
   for(i = 0; i < PLA_NUMBER_OF_STRUCTS; i++) /* number of structs == number of views */
      if(PLA_GLOBAL_VIEWS_INDEX[i] == 0)
      {
         index = i;
         PLA_GLOBAL_VIEWS_INDEX[index] = 1; /* used */
         break;
      }
   return(index);
}

int PLA_Initialized_object_structs(void)
{
   return(PLA_MEMORY_SET_ASIDE);
}

