#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

extern nb;
 
/* LAPACK eigensolver */
void dsyev_(char *,char *,int *,double *,int *,double *,double *,int *,int *);

void prob_to_node( PRISM_subproblem prob );
void gather_to_node( PRISM_subproblem prob );
void local_solve( PRISM_subproblem prob );
void update_global( PRISM_subproblem prob );
 
void PRISM_end_game(PRISM_final_list final_list)
{
/* 
   Complete subproblems from final_list.
*/
   PRISM_subproblem temp;

   /* Redistribute subproblem */
   PRISM_final_list_get_first( final_list, &temp );
   while (temp) {
       prob_to_node( temp );
       gather_to_node( temp );
       PRISM_final_list_get_next( final_list, &temp );
   }

   /* Solve using LAPACK - need to add option for PRISM */
   PRISM_final_list_get_first( final_list, &temp );
   while (temp) {
       local_solve( temp );
       PRISM_final_list_get_next( final_list, &temp );
   }

   /* Update global vectors with local solution */ 
   PRISM_final_list_get_first( final_list, &temp );
   while (temp) {
       update_global( temp );
       PRISM_final_list_get_next( final_list, &temp );
   }

#ifdef DB_GLOBAL_
   if (me==ROOT) printf(">> PRISM_end_game\n");
#endif
}


/* Determine which node to send subproblem */
void prob_to_node( PRISM_subproblem prob ) {
   int numnodes;
   PLA_Template template;

   PLA_Obj_template( prob->A, &template );
   PLA_Temp_comm_all_size( template, &numnodes );
   prob->node = prob->id % numnodes; 
}


/* Gather subproblem to node */
void gather_to_node( PRISM_subproblem prob ) {
   int col_size, owner_row, owner_col;
   PLA_Template template;

   /* Determine location in mesh for local subproblem */
   PLA_Obj_template( prob->A, &template );
   PLA_Temp_comm_col_size( template, &col_size );

   /* need to verify that this works in all cases */
   owner_row = prob->node % col_size;
   owner_col = prob->node / col_size;

   /* Create multiscalars to hold local subproblem */
   /* This crashes or hangs for some template block sizes - why? */ 
   PLA_Mscalar_create_conf_to( prob->A, owner_row, owner_col, &prob->sub );
   PLA_Mscalar_create_conf_to( prob->D, owner_row, owner_col, &prob->diag );

   /* Gather subproblem to one node */
   /* This crashes or hangs for some template block sizes - why? */ 
   PLA_Copy( prob->A, prob->sub );
}

/* Solve local subproblem with LAPACK */
void local_solve( PRISM_subproblem prob ) {
   int i, lwork, info;
   int rank, length, ldim;
   double *work;
   void *sub, *diag;
   char jobz='V', uplo='U';
   PLA_Template template;

   PLA_Obj_template( prob->sub, &template );
   PLA_Temp_comm_all_rank( template, &rank );

   if ( rank != prob->node )
      return;

   /* Get info about local subproblem */
   PLA_Obj_local_ldim    ( prob->sub, &ldim );
   PLA_Obj_local_length  ( prob->sub, &length );
   PLA_Obj_local_buffer  ( prob->sub, &sub ); 
   PLA_Obj_local_buffer  ( prob->diag, &diag ); 

   /* Get workspace for LAPACK function */
   lwork = (nb+2) * length;
   work = (double *) malloc( sizeof(double)*lwork );
   if ( work == NULL ) {
      printf("malloc error in local_solve\n");
      exit(1);
   }
   /* This fixes a bug on the UW SP2 */
   for (i=0; i<lwork; i++) work[i] = 0.; 

   /* Call LAPACK eigensolver */
   dsyev_( &jobz, &uplo, &length, sub, &ldim, diag, work, &lwork, &info ); 
   if (info != 0) {
      printf("LAPACK dsyev failed, info=%d\n",info); 
      exit(1);
   }
   free(work);
}


/* Update global vectors based on the local solution */
void update_global( PRISM_subproblem prob ) {
   PLA_Obj Zhat=NULL;

   /* Update global eigenvectors */
   PLA_Matrix_create_conf_to( prob->Z, &Zhat );
   PLA_Obj_set_to_zero( Zhat );
   PLA_Copy( prob->sub, prob->A );

   PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, prob->Z, 
       prob->A, PRISM_zero, Zhat );
   PLA_Local_copy( Zhat, prob->Z );

   PLA_Obj_free( &Zhat );

   /* Update global eigenvalues */
   PLA_Copy( prob->diag, prob->D );
}
