#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"


void PRISM_scale_shift(PLA_Obj X, PLA_Obj lambda_min, PLA_Obj lambda_mean,
			     PLA_Obj lambda_max) {
/*
    Scale and shift X so that all eigenvalues are in the interval
    [ 0,1 ] and the mean goes to 1/2. 

    Method:
    lambda_ave = (lambda_min + lambda_max) / 2.
    lambda_use = lambda_mean <= lambda_ave ? lambda_max : lambda_min
    X = ( (X - lambda_mean*I) / (lambda_use - lambda_mean) + I ) / 2.
*/
    double        ave, use, temp;
    double        temp_min, temp_max, temp_mean;
    void          *lambda_min_buf, *lambda_max_buf, *lambda_mean_buf;
    PLA_Obj       alpha=NULL;
    MPI_Datatype  datatype;
    PLA_Template  template;

    /* Extract global info from matrix X */
    PLA_Obj_datatype( X, &datatype );
    PLA_Obj_template( X, &template );

    /* Scale factor for PLA_Scal operations */
    PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 
       1, 1, template, &alpha );

    /* Extract local info from lambda_min, _max, and _mean  */
    PLA_Obj_local_buffer( lambda_min,  &lambda_min_buf );
    PLA_Obj_local_buffer( lambda_max,  &lambda_max_buf );
    PLA_Obj_local_buffer( lambda_mean, &lambda_mean_buf );

    temp_min =  *(double *)lambda_min_buf;
    temp_max =  *(double *)lambda_max_buf;
    temp_mean = *(double *)lambda_mean_buf;

    /* Compute lambda_ave and lambda_use */
    ave = .5 * ( temp_min + temp_max );
    use = ( temp_mean <= ave ? temp_max : temp_min );

    /* Subtract lambda_mean from matrix X */
    temp = -temp_mean;
    PLA_Obj_set( alpha, datatype, &temp );
    PRISM_add_to_diag( alpha, X );

    /* Scale result by 1./(lambda_use-lambda_mean) */
    temp = .5 / (use - temp_mean);
    PLA_Obj_set( alpha, datatype, &temp );
    PLA_Local_scal( alpha, X ); 

    /* Second matrix addition */
    temp = .5;
    PLA_Obj_set( alpha, datatype, &temp );
    PRISM_add_to_diag( alpha, X );

    /* free objects */
    PLA_Obj_free(&alpha);

#ifdef DB_GLOBAL_
    if (me==ROOT) printf(">> PRISM_scale_shift\n"); 
    PRISM_print(X);
#endif
}


