#include <stdio.h>
#include <math.h>

#define fifi2mod(x,y)  ((x)%(y))

/*
  THIS PACKAGE HAS BEEN CHANGED TO DOUBLE PRECISION BY L.SHTILMAN, 1987
 PURPOSE      PERFORMS MULTIPLE FAST FOURIER TRANSFORMS.  THIS PACKAGE
              WILL PERFORM A NUMBER OF SIMULTANEOUS COMPLEX PERIODIC
              FOURIER TRANSFORMS OR CORRESPONDING INVERSE TRANSFORMS.
              THAT IS, GIVEN A SET OF COMPLEX GRIDPOINT VECTORS, THE
              PACKAGE RETURNS A SET OF COMPLEX FOURIER
              COEFFICIENT VECTORS, OR VICE VERSA.  THE LENGTH OF THE
              TRANSFORMS MUST BE A NUMBER GREATER THAN 1 THAT HAS
              NO PRIME FACTORS OTHER THAN 2, 3, AND 5.
              THE PACKAGE CFFT99 CONTAINS SEVERAL USER-LEVEL ROUTINES:
            SUBROUTINE CFTFAX
                AN INITIALIZATION ROUTINE THAT MUST BE CALLED ONCE
                BEFORE A SEQUENCE OF CALLS TO CFFT99
                (PROVIDED THAT N IS NOT CHANGED).
            SUBROUTINE CFFT99
                THE ACTUAL TRANSFORM ROUTINE ROUTINE, CABABLE OF
                PERFORMING BOTH THE TRANSFORM AND ITS INVERSE.
                HOWEVER, AS THE TRANSFORMS ARE NOT NORMALIZED,
                THE APPLICATION OF A TRANSFORM FOLLOWED BY ITS
                INVERSE WILL YIELD THE ORIGINAL VALUES MULTIPLIED
                BY N.
 ACCESS       *FORTRAN,P=XLIB,SN=CFFT99
 USAGE        LET N BE OF THE FORM 2**P * 3**Q * 5**R, WHERE P .GE. 0,
              Q .GE. 0, AND R .GE. 0.  THEN A TYPICAL SEQUENCE OF
              CALLS TO TRANSFORM A GIVEN SET OF COMPLEX VECTORS OF
              LENGTH N TO A SET OF (UNSCALED) COMPLEX FOURIER
              COEFFICIENT VECTORS OF LENGTH N IS
                   DIMENSION IFAX(13),TRIGS(2*N)
                   COMPLEX A(...), WORK(...)
                   CALL CFTFAX (N, IFAX, TRIGS)
                   CALL CFFT99 (A,WORK,TRIGS,IFAX,INC,JUMP,N,LOT,ISIGN)
              THE OUTPUT VECTORS OVERWRITE THE INPUT VECTORS, AND
              THESE ARE STORED IN A.  WITH APPROPRIATE CHOICES FOR
              THE OTHER ARGUMENTS, THESE VECTORS MAY BE CONSIDERED
              EITHER THE ROWS OR THE COLUMNS OF THE ARRAY A.
              SEE THE INDIVIDUAL WRITE-UPS FOR CFTFAX AND
              CFFT99 BELOW, FOR A DETAILED DESCRIPTION OF THE
              ARGUMENTS.
 HISTORY      THE PACKAGE WAS WRITTEN BY CLIVE TEMPERTON AT ECMWF IN
              NOVEMBER, 1978.  IT WAS MODIFIED, DOCUMENTED, AND TESTED
              FOR NCAR BY RUSS REW IN SEPTEMBER, 1980.  IT WAS
              FURTHER MODIFIED FOR THE FULLY COMPLEX CASE BY DAVE
              FULKER IN NOVEMBER, 1980.
-----------------------------------------------------------------------
 SUBROUTINE CFTFAX (N,IFAX,TRIGS)
 PURPOSE      A SET-UP ROUTINE FOR CFFT99.  IT NEED ONLY BE
              CALLED ONCE BEFORE A SEQUENCE OF CALLS TO CFFT99,
              PROVIDED THAT N IS NOT CHANGED.
 ARGUMENT     IFAX(13),TRIGS(2*N)
 DIMENSIONS
 ARGUMENTS
 ON INPUT     N
               AN EVEN NUMBER GREATER THAN 1 THAT HAS NO PRIME FACTOR
               GREATER THAN 5.  N IS THE LENGTH OF THE TRANSFORMS (SEE
               THE DOCUMENTATION FOR CFFT99 FOR THE DEFINITION OF
               THE TRANSFORMS).
              IFAX
               AN INTEGER ARRAY.  THE NUMBER OF ELEMENTS ACTUALLY USED
               WILL DEPEND ON THE FACTORIZATION OF N.  DIMENSIONING
               IFAX FOR 13 SUFFICES FOR ALL N LESS THAN 1 MILLION.
              TRIGS
               A REAL ARRAY OF DIMENSION 2*N
 ON OUTPUT    IFAX
               CONTAINS THE FACTORIZATION OF N.  IFAX(1) IS THE
               NUMBER OF FACTORS, AND THE FACTORS THEMSELVES ARE STORED
               IN IFAX(2),IFAX(3),...  IF N HAS ANY PRIME FACTORS
               GREATER THAN 5, IFAX(1) IS SET TO -99.
              TRIGS
               AN ARRAY OF TRIGONOMETRIC FUNCTION VALUES SUBSEQUENTLY
               USED BY THE CFT ROUTINES.
-----------------------------------------------------------------------
 SUBROUTINE CFFT99 (A,WORK,TRIGS,IFAX,INC,JUMP,N,LOT,ISIGN)
 PURPOSE      PERFORM A NUMBER OF SIMULTANEOUS (UNNORMALIZED) COMPLEX
              PERIODIC FOURIER TRANSFORMS OR CORRESPONDING INVERSE
              TRANSFORMS.  GIVEN A SET OF COMPLEX GRIDPOINT
              VECTORS, THE PACKAGE RETURNS A SET OF
              COMPLEX FOURIER COEFFICIENT VECTORS, OR VICE
              VERSA.  THE LENGTH OF THE TRANSFORMS MUST BE A
              NUMBER HAVING NO PRIME FACTORS OTHER THAN
              2, 3, AND 5.  THIS ROUTINE IS
              OPTIMIZED FOR USE ON THE CRAY-1.
 ARGUMENT     COMPLEX A(N*INC+(LOT-1)*JUMP), WORK(N*LOT)
 DIMENSIONS   REAL TRIGS(2*N), INTEGER IFAX(13)
 ARGUMENTS
 ON INPUT     A
               A COMPLEX ARRAY OF LENGTH N*INC+(LOT-1)*JUMP CONTAINING
               THE INPUT GRIDPOINT OR COEFFICIENT VECTORS.  THIS ARRAY
               OVERWRITTEN BY THE RESULTS.
               N.B. ALTHOUGH THE ARRAY A IS USUALLY CONSIDERED TO BE OF
               TYPE COMPLEX IN THE CALLING PROGRAM, IT IS TREATED AS
               REAL WITHIN THE TRANSFORM PACKAGE.  THIS REQUIRES THAT
               SUCH TYPE CONFLICTS ARE PERMITTED IN THE USER^S
               ENVIRONMENT, AND THAT THE STORAGE OF COMPLEX NUMBERS
               MATCHES THE ASSUMPTIONS OF THIS ROUTINE.  THIS ROUTINE
               ASSUMES THAT THE REAL AND IMAGINARY PORTIONS OF A
               COMPLEX NUMBER OCCUPY ADJACENT ELEMENTS OF MEMORY.  IF
               THESE CONDITIONS ARE NOT MET, THE USER MUST TREAT THE
               ARRAY A AS REAL (AND OF TWICE THE ABOVE LENGTH), AND
               WRITE THE CALLING PROGRAM TO TREAT THE REAL AND
               IMAGINARY PORTIONS EXPLICITLY.
              WORK
               A COMPLEX WORK ARRAY OF LENGTH N*LOT OR A REAL ARRAY
               OF LENGTH 2*N*LOT.  SEE N.B. ABOVE.
              TRIGS
               AN ARRAY SET UP BY CFTFAX, WHICH MUST BE CALLED FIRST.
              IFAX
               AN ARRAY SET UP BY CFTFAX, WHICH MUST BE CALLED FIRST.
               N.B. IN THE FOLLOWING ARGUMENTS, INCREMENTS ARE MEASURED
               IN WORD PAIRS, BECAUSE EACH COMPLEX ELEMENT IS ASSUMED
               TO OCCUPY AN ADJACENT PAIR OF WORDS IN MEMORY.
              INC
               THE INCREMENT (IN WORD PAIRS) BETWEEN SUCCESSIVE ELEMENT
               OF EACH (COMPLEX) GRIDPOINT OR COEFFICIENT VECTOR
               (E.G.  INC=1 FOR CONSECUTIVELY STORED DATA).
              JUMP
               THE INCREMENT (IN WORD PAIRS) BETWEEN THE FIRST ELEMENTS
               OF SUCCESSIVE DATA OR COEFFICIENT VECTORS.  ON THE CRAY-
               TRY TO ARRANGE DATA SO THAT JUMP IS NOT A MULTIPLE OF 8
               (TO AVOID MEMORY BANK CONFLICTS).  FOR CLARIFICATION OF
               INC AND JUMP, SEE THE EXAMPLES BELOW.
              N
               THE LENGTH OF EACH TRANSFORM (SEE DEFINITION OF
               TRANSFORMS, BELOW).
              LOT
               THE NUMBER OF TRANSFORMS TO BE DONE SIMULTANEOUSLY.
              ISIGN
               = -1 FOR A TRANSFORM FROM GRIDPOINT VALUES TO FOURIER
                    COEFFICIENTS.
               = +1 FOR A TRANSFORM FROM FOURIER COEFFICIENTS TO
                    GRIDPOINT VALUES.
 ON OUTPUT    A
               IF ISIGN = -1, AND LOT GRIDPOINT VECTORS ARE SUPPLIED,
               EACH CONTAINING THE COMPLEX SEQUENCE:
               G(0),G(1), ... ,G(N-1)  (N COMPLEX VALUES)
               THEN THE RESULT CONSISTS OF LOT COMPLEX VECTORS EACH
               CONTAINING THE CORRESPONDING N COEFFICIENT VALUES:
               C(0),C(1), ... ,C(N-1)  (N COMPLEX VALUES)
               DEFINED BY:
                 C(K) = SUM(J=0,...,N-1)( G(J)*EXP(-2*I*J*K*PI/N) )
                 WHERE I = SQRT(-1)
               IF ISIGN = +1, AND LOT COEFFICIENT VECTORS ARE SUPPLIED,
               EACH CONTAINING THE COMPLEX SEQUENCE:
               C(0),C(1), ... ,C(N-1)  (N COMPLEX VALUES)
               THEN THE RESULT CONSISTS OF LOT COMPLEX VECTORS EACH
               CONTAINING THE CORRESPONDING N GRIDPOINT VALUES:
               G(0),G(1), ... ,G(N-1)  (N COMPLEX VALUES)
               DEFINED BY:
                 G(J) = SUM(K=0,...,N-1)( G(K)*EXP(+2*I*J*K*PI/N) )
                 WHERE I = DSQRT(-1)
               A CALL WITH ISIGN=-1 FOLLOWED BY A CALL WITH ISIGN=+1
               (OR VICE VERSA) RETURNS THE ORIGINAL DATA, MULTIPLIED
               BY THE FACTOR N.
 EXAMPLE       GIVEN A 64 BY 9 GRID OF COMPLEX VALUES, STORED IN
               A 66 BY 9 COMPLEX ARRAY, A, COMPUTE THE TWO DIMENSIONAL
               FOURIER TRANSFORM OF THE GRID.  FROM TRANSFORM THEORY,
               IT IS KNOWN THAT A TWO DIMENSIONAL TRANSFORM CAN BE
               OBTAINED BY FIRST TRANSFORMING THE GRID ALONG ONE
               DIRECTION, THEN TRANSFORMING THESE RESULTS ALONG THE
               ORTHOGONAL DIRECTION.
               COMPLEX A(66,9), WORK(64,9)
               REAL TRIGS1(128), TRIGS2(18)
               INTEGER IFAX1(13), IFAX2(13)
               SET UP THE IFAX AND TRIGS ARRAYS FOR EACH DIRECTION:
               CALL CFTFAX(64, IFAX1, TRIGS1)
               CALL CFTFAX( 9, IFAX2, TRIGS2)
               IN THIS CASE, THE COMPLEX VALUES OF THE GRID ARE
               STORED IN MEMORY AS FOLLOWS (USING U AND V TO
               DENOTE THE REAL AND IMAGINARY COMPONENTS, AND
               ASSUMING CONVENTIONAL FORTRAN STORAGE):
   U(1,1), V(1,1), U(2,1), V(2,1),  ...  U(64,1), V(64,1), 4 NULLS,
   U(1,2), V(1,2), U(2,2), V(2,2),  ...  U(64,2), V(64,2), 4 NULLS,
   .       .       .       .         .   .        .        .
   .       .       .       .         .   .        .        .
   .       .       .       .         .   .        .        .
   U(1,9), V(1,9), U(2,9), V(2,9),  ...  U(64,9), V(64,9), 4 NULLS.
               WE CHOOSE (ARBITRARILY) TO TRANSORM FIRST ALONG THE
               DIRECTION OF THE FIRST SUBSCRIPT.  THUS WE DEFINE
               THE LENGTH OF THE TRANSFORMS, N, TO BE 64, THE
               NUMBER OF TRANSFORMS, LOT, TO BE 9, THE INCREMENT
               BETWEEN ELEMENTS OF EACH TRANSFORM, INC, TO BE 1,
               AND THE INCREMENT BETWEEN THE STARTING POINTS
               FOR EACH TRANSFORM, JUMP, TO BE 66 (THE FIRST
               DIMENSION OF A).
               CALL CFFT99( A, WORK, TRIGS1, IFAX1, 1, 66, 64, 9, -1)
               TO TRANSFORM ALONG THE DIRECTION OF THE SECOND SUBSCRIPT
               THE ROLES OF THE INCREMENTS ARE REVERSED.  THUS WE DEFIN
               THE LENGTH OF THE TRANSFORMS, N, TO BE 9, THE
               NUMBER OF TRANSFORMS, LOT, TO BE 64, THE INCREMENT
               BETWEEN ELEMENTS OF EACH TRANSFORM, INC, TO BE 66,
               AND THE INCREMENT BETWEEN THE STARTING POINTS
               FOR EACH TRANSFORM, JUMP, TO BE 1
               CALL CFFT99( A, WORK, TRIGS2, IFAX2, 66, 1, 9, 64, -1)
               THESE TWO SEQUENTIAL STEPS RESULTS IN THE TWO-DIMENSIONA
               FOURIER COEFFICIENT ARRAY OVERWRITING THE INPUT
               GRIDPOINT ARRAY, A.  THE SAME TWO STEPS APPLIED AGAIN
               WITH ISIGN = +1 WOULD RESULT IN THE RECONSTRUCTION OF
               THE GRIDPOINT ARRAY (MULTIPLIED BY A FACTOR OF 64*9).
-----------------------------------------------------------------------
*/
void cfft99(a, work, trigs, ifax, inc, jump, n, lot, isign)
int ifax[],*inc,*jump,*n,*lot,*isign;
float a[],work[],trigs[];
{
  extern void vpassm();
  static int K1 = 2;
  static int nn,ink,jum,nfax,jnk,jst,ibase,
    ilast,nh,l,i1,i2,m,jbase,i,j,igo,la,k;
  static float hreal,himag;
  /*
     SUBROUTINE ^CFFT99^ - MULTIPLE FAST COMPLEX FOURIER TRANSFORM
     A IS THE ARRAY CONTAINING INPUT AND OUTPUT DATA
     WORK IS AN AREA OF SIZE N*LOT
     TRIGS IS A PREVIOUSLY PREPARED LIST OF TRIG FUNCTION VALUES
     IFAX IS A PREVIOUSLY PREPARED LIST OF FACTORS OF N
     INC IS THE INCREMENT WITHIN EACH DATA #VECTOR#
         (E.G. INC=1 FOR CONSECUTIVELY STORED DATA)
     JUMP IS THE INCREMENT BETWEEN THE START OF EACH DATA VECTOR
     N IS THE LENGTH OF THE DATA VECTORS
     LOT IS THE NUMBER OF DATA VECTORS
     ISIGN = +1 FOR TRANSFORM FROM SPECTRAL TO GRIDPOINT
           = -1 FOR TRANSFORM FROM GRIDPOINT TO SPECTRAL
     VECTORIZATION IS ACHIEVED ON CRAY BY DOING THE TRANSFORMS IN
     PARALLEL.
*/

  nn = *n+*n;
  ink = *inc+*inc;
  jum = *jump+*jump;
  nfax = ifax[0];
  jnk = jst = 2;
  if(*isign >= 0) goto S30;

/*
     THE INNERMOST TEMPERTON ROUTINES HAVE NO FACILITY FOR THE
     FORWARD (ISIGN = -1) TRANSFORM.  THEREFORE, THE INPUT MUST BE
     REARRANGED AS FOLLOWS:
     THE ORDER OF EACH INPUT VECTOR,
     G(0), G(1), G(2), ... , G(N-2), G(N-1)
     IS REVERSED (EXCLUDING G(0)) TO YIELD
     G(0), G(N-1), G(N-2), ... , G(2), G(1).
     WITHIN THE TRANSFORM, THE CORRESPONDING EXPONENTIAL MULTIPLIER
     IS THEN PRECISELY THE CONJUGATE OF THAT FOR THE NORMAL
     ORDERING.  THUS THE FORWARD (ISIGN = -1) TRANSFORM IS
     ACCOMPLISHED
     FOR NFAX ODD, THE INPUT MUST BE TRANSFERRED TO THE WORK ARRAY,
     AND THE REARRANGEMENT CAN BE DONE DURING THE MOVE.
*/

  jnk = -2;
  jst = nn-2;
  if(fifi2mod(nfax,2) == 1) goto S40;

/*
     FOR NFAX EVEN, THE REARRANGEMENT MUST BE APPLIED DIRECTLY TO
     THE INPUT ARRAY.  THIS CAN BE DONE BY SWAPPING ELEMENTS.
*/

  ibase = 1;
  ilast = (*n-1)*ink;
  nh = *n/2;
  for(l=1; l<=*lot; l++) {
    i1 = ibase+ink;
    i2 = ibase+ilast;
    
    for(m=1; m<=nh; m++) {
/*
     SWAP REAL AND IMAGINARY PORTIONS
*/
      hreal = a[i1-1];
      himag = a[i1];
      a[i1-1] = a[i2-1];
      a[i1] = a[i2];
      a[i2-1] = hreal;
      a[i2] = himag;
      i1 += ink;
      i2 -= ink;
    }
    ibase += jum;
  }
  goto S100;
S30:
  if(fifi2mod(nfax,2) == 0) goto S100;
S40:
/*
     DURING THE TRANSFORM PROCESS, NFAX STEPS ARE TAKEN, AND THE
     RESULTS ARE STORED ALTERNATELY IN WORK AND IN A.  IF NFAX IS
     ODD, THE INPUT DATA ARE FIRST MOVED TO WORK SO THAT THE FINAL
     RESULT (AFTER NFAX STEPS) IS STORED IN ARRAY A.
*/

  ibase = jbase = 1;
  for(l=1; l<=*lot; l++) {
/*
     MOVE REAL AND IMAGINARY PORTIONS OF ELEMENT ZERO
*/
    work[jbase-1] = a[ibase-1];
    work[jbase] = a[ibase];
    i = ibase+ink;
    j = jbase+jst;
/*
  DIR$ IVDEP
*/
    for(m=2; m<=*n; m++) {
/*
     MOVE REAL AND IMAGINARY PORTIONS OF OTHER ELEMENTS (POSSIBLY IN
     REVERSE ORDER, DEPENDING ON JST AND JNK)
*/

      work[j-1] = a[i-1];
      work[j] = a[i];
      i += ink;
      j += jnk;
    }
    ibase += jum;
    jbase += nn;
  }
S100:
/*
     PERFORM THE TRANSFORM PASSES, ONE PASS FOR EACH FACTOR.  DURING
     EACH PASS THE DATA ARE MOVED FROM A TO WORK OR FROM WORK TO A.
     FOR NFAX EVEN, THE FIRST PASS MOVES FROM A TO WORK
*/

  igo = 110;

/*
     FOR NFAX ODD, THE FIRST PASS MOVES FROM WORK TO A
*/

  if(fifi2mod(nfax,2) == 1) igo = 120;
  la = 1;
  for(k=1; k<=nfax; k++) {
    if(igo == 120) goto S120;
    vpassm(&a[0],&a[1],&work[0],&work[1],trigs,&ink,&K1,&jum,&nn,lot,n,&
	   ifax[k],&la);
    igo = 120;
    goto S130;
  S120:
    vpassm(&work[0],&work[1],&a[0],&a[1],trigs,&K1,&ink,&nn,&jum,lot,n,&
	   ifax[k],&la);
    igo = 110;
  S130:
    la *= ifax[k];
  }

/*
     AT THIS POINT THE FINAL TRANSFORM RESULT IS STORED IN A.
*/

  return;
}


/*************************************************************************
* Function: cftfax()
*
**************************************************************************/
void cftfax(n, ifax, trigs)
/*
     THIS ROUTINE WAS MODIFIED FROM TEMPERTON^S ORIGINAL
     BY DAVE FULKER.  IT NO LONGER PRODUCES FACTORS IN ASCENDING
     ORDER, AND THERE ARE NONE OF THE ORIGINAL #MODE# OPTIONS.
 ON INPUT     N
               THE LENGTH OF EACH COMPLEX TRANSFORM TO BE PERFORMED
               N MUST BE GREATER THAN 1 AND CONTAIN NO PRIME
               FACTORS GREATER THAN 5.
 ON OUTPUT    IFAX
               IFAX(1)
                 THE NUMBER OF FACTORS CHOSEN OR -99 IN CASE OF ERROR
               IFAX(2) THRU IFAX( IFAX(1)+1 )
                 THE FACTORS OF N IN THE FOLLOWIN ORDER:  APPEARING
                 FIRST ARE AS MANY FACTORS OF 4 AS CAN BE OBTAINED.
                 SUBSEQUENT FACTORS ARE PRIMES, AND APPEAR IN
                 ASCENDING ORDER, EXCEPT FOR MULTIPLE FACTORS.
              TRIGS
               2N SIN AND COS VALUES FOR USE BY THE TRANSFORM ROUTINE
*/

int *n,ifax[];
float trigs[];
{
  extern void fact(),cftrig();
  static int k;

  fact(n,ifax);
  k = ifax[0];
  if(k < 1 || ifax[k] > 5) ifax[0] = -99;
  if(ifax[0] <= 0) puts(" FFTFAX -- INVALID N");
  cftrig(n,trigs);
  return;
}


/****************************************************************************
* Function: fact()
*
******************************************************************************/
void fact(n,ifax)
/*
     FACTORIZATION ROUTINE THAT FIRST EXTRACTS ALL FACTORS OF 4
*/
int *n,ifax[];
{
  static int nn,k,l,max,inc;

  if(*n > 1) goto S10;
  ifax[0] = 0;
  if(*n < 1) ifax[0] = -99;
  return;
S10:
  nn = *n;
  k = 1;
S20:

/*
     TEST FOR FACTORS OF 4
*/

  if(fifi2mod(nn,4) != 0) goto S30;
  k += 1;
  ifax[k-1] = 4;
  nn /= 4;
  if(nn == 1) goto S80;
  goto S20;
S30:

/*
     TEST FOR EXTRA FACTOR OF 2
*/

  if(fifi2mod(nn,2) != 0) goto S40;
  k += 1;
  ifax[k-1] = 2;
  nn /= 2;
  if(nn == 1) goto S80;
S40:

/*
     TEST FOR FACTORS OF 3
*/

  if(fifi2mod(nn,3) != 0) goto S50;
  k += 1;
  ifax[k-1] = 3;
  nn /= 3;
  if(nn == 1) goto S80;
  goto S40;
S50:

/*
     NOW FIND REMAINING FACTORS
*/

  l = 5;
  max = sqrt((float)nn);
  inc = 2;
S60:

/*
     INC ALTERNATELY TAKES ON VALUES 2 AND 4
*/

  if(fifi2mod(nn,l) != 0) goto S70;
  k += 1;
  ifax[k-1] = l;
  nn /= l;
  if(nn == 1) goto S80;
  goto S60;
S70:
  if(l > max) goto S75;
  l += inc;
  inc = 6-inc;
  goto S60;
S75:
  k += 1;
  ifax[k-1] = nn;
S80:
  ifax[0] = k-1;

/*
     IFAX(1) NOW CONTAINS NUMBER OF FACTORS
*/

  return;
}


/**************************************************************************
* Function: cftrig()
*
***************************************************************************/
void cftrig(n,trigs)
int *n;
float trigs[];
{
  static int l,i;
  static float pi,del,angle;

  pi = 2.0*asin(1.0);
  del = (pi+pi)/(float)*n;
  l = *n+*n;
  for(i=1; i<=l; i+=2) {
    angle = 0.5*(float)(i-1)*del;
    trigs[i-1] = cos(angle);
    trigs[i] = sin(angle);
  }
  return;
}

void vpassm1(a, b, c, d, trigs, inc1, inc2, inc3, inc4, lot, n, ifac, la,
             m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
             D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,c1,s1,c2,s2,
             c3,s3,c4,s4)
int *inc1,*inc2,*inc3,*inc4,*lot,*n,*ifac,*la;
float a[],b[],c[],d[],trigs[];
int m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,D1,D2;
int ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8;
float c1,s1,c2,s2,c3,s3,c4,s4;
{
  static float sin36 = 0.587785252292473;
  static float cos36 = 0.809016994374947;
  static float sin72 = 0.951056516295154;
  static float cos72 = 0.309016994374947;
  static float sin60 = 0.866025403784437;

/*
     CODING FOR FACTOR 2
*/

  ia = ja = 1;
  ib = ia+iink;
  jb = ja+jink;
  for(l=1; l<=*la; l++) {
    i = ibase;
    j = jbase;

/*
  DIR$ IVDEP
*/

    for(ijk=1; ijk<=*lot; ijk++) {
      c[ja+j-1] = a[ia+i-1]+a[ib+i-1];
      d[ja+j-1] = b[ia+i-1]+b[ib+i-1];
      c[jb+j-1] = a[ia+i-1]-a[ib+i-1];
      d[jb+j-1] = b[ia+i-1]-b[ib+i-1];
      i += *inc3;
      j += *inc4;
    }
    ibase += *inc1;
    jbase += *inc2;
  }
  if(*la == m) return;
  la1 = *la+1;
  jbase += jump;
  for(k=la1,D1= *la,D2=(m-k+D1)/D1; D2>0; D2--,k+=D1) {
    kb = k+k-2;
    c1 = trigs[kb];
    s1 = trigs[kb+1];
    for(l=1; l<=*la; l++) {
      i = ibase;
      j = jbase;

/*
  DIR$ IVDEP
*/

      for(ijk=1; ijk<=*lot; ijk++) {
	c[ja+j-1] = a[ia+i-1]+a[ib+i-1];
	d[ja+j-1] = b[ia+i-1]+b[ib+i-1];
	c[jb+j-1] = c1*(a[ia+i-1]-a[ib+i-1])-s1*(b[ia+i-1]-b[ib+i-1]);
	d[jb+j-1] = s1*(a[ia+i-1]-a[ib+i-1])+c1*(b[ia+i-1]-b[ib+i-1]);
	i += *inc3;
	j += *inc4;
      }
      ibase += *inc1;
      jbase += *inc2;
    }
    jbase += jump;
  }
  return;
}

void vpassm2(a, b, c, d, trigs, inc1, inc2, inc3, inc4, lot, n, ifac, la,
             m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
             D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,c1,s1,c2,s2,
             c3,s3,c4,s4)
int *inc1,*inc2,*inc3,*inc4,*lot,*n,*ifac,*la;
float a[],b[],c[],d[],trigs[];
int m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,D1,D2;
int ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8;
float c1,s1,c2,s2,c3,s3,c4,s4;
{
  static float sin36 = 0.587785252292473;
  static float cos36 = 0.809016994374947;
  static float sin72 = 0.951056516295154;
  static float cos72 = 0.309016994374947;
  static float sin60 = 0.866025403784437;

/*
     CODING FOR FACTOR 3
*/

  ia = ja = 1;
  ib = ia+iink;
  jb = ja+jink;
  ic = ib+iink;
  jc = jb+jink;
  for(l=1; l<=*la; l++) {
    i = ibase;
    j = jbase;

/*
  DIR$ IVDEP
*/

    for(ijk=1; ijk<=*lot; ijk++) {
      c[ja+j-1] = a[ia+i-1]+(a[ib+i-1]+a[ic+i-1]);
      d[ja+j-1] = b[ia+i-1]+(b[ib+i-1]+b[ic+i-1]);
      c[jb+j-1] = a[ia+i-1]-0.5*(a[ib+i-1]+a[ic+i-1])-sin60*(b[ib+i-1]-b[ic+i-1]);
      c[jc+j-1] = a[ia+i-1]-0.5*(a[ib+i-1]+a[ic+i-1])+sin60*(b[ib+i-1]-b[ic+i-1]);
      d[jb+j-1] = b[ia+i-1]-0.5*(b[ib+i-1]+b[ic+i-1])+sin60*(a[ib+i-1]-a[ic+i-1]);
      d[jc+j-1] = b[ia+i-1]-0.5*(b[ib+i-1]+b[ic+i-1])-sin60*(a[ib+i-1]-a[ic+i-1]);
      i += *inc3;
      j += *inc4;
    }
    ibase += *inc1;
    jbase += *inc2;
  }
  if(*la == m) 
    return;
  la1 = *la+1;
  jbase += jump;
  for(k=la1,D3= *la,D4=(m-k+D3)/D3; D4>0; D4--,k+=D3) {
    kb = k+k-2;
    kc = kb+kb;
    c1 = trigs[kb];
    s1 = trigs[kb+1];
    c2 = trigs[kc];
    s2 = trigs[kc+1];
    for(l=1; l<=*la; l++) {
      i = ibase;
      j = jbase;

/*
  DIR$ IVDEP
*/

      for(ijk=1; ijk<=*lot; ijk++) {
	c[ja+j-1] = a[ia+i-1]+(a[ib+i-1]+a[ic+i-1]);
	d[ja+j-1] = b[ia+i-1]+(b[ib+i-1]+b[ic+i-1]);
	c[jb+j-1] = c1*(a[ia+i-1]-0.5*(a[ib+i-1]+a[ic+i-1])-
			sin60*(b[ib+i-1]-b[ic+i-1]))-
	  s1*(b[ia+i-1]-0.5*(b[ib+i-1]+b[ic+i-1])+sin60*(a[ib+i-1]-a[ic+i-1]));
	d[jb+j-1] = s1*(a[ia+i-1]-0.5*(a[ib+i-1]+a[ic+i-1])-
			sin60*(b[ib+i-1]-b[ic+i-1]))+c1*(b[ia+i-1]-0.5*
							 (b[ib+i-1]+b[ic+i-1])+
							 sin60*(a[ib+i-1]-a[ic+i-1]));
	c[jc+j-1] = c2*(a[ia+i-1]-0.5*(a[ib+i-1]+a[ic+i-1])+sin60*(b[ib+
                  i-1]-b[ic+i-1]))-s2*(b[ia+i-1]-0.5*(b[ib+i-1]+b[ic+i-1])-
                  sin60*(a[ib+i-1]-a[ic+i-1]));
	d[jc+j-1] = s2*(a[ia+i-1]-0.5*(a[ib+i-1]+a[ic+i-1])+sin60*(b[ib+
                  i-1]-b[ic+i-1]))+c2*(b[ia+i-1]-0.5*(b[ib+i-1]+b[ic+i-1])-
                  sin60*(a[ib+i-1]-a[ic+i-1]));
	i += *inc3;
	j += *inc4;
      }
      ibase += *inc1;
      jbase += *inc2;
    }
    jbase += jump;
  }
  return;
}

void vpassm3(a, b, c, d, trigs, inc1, inc2, inc3, inc4, lot, n, ifac, la,
             m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
             D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,c1,s1,c2,s2,
             c3,s3,c4,s4)
int *inc1,*inc2,*inc3,*inc4,*lot,*n,*ifac,*la;
float a[],b[],c[],d[],trigs[];
int m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,D1,D2;
int ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8;
float c1,s1,c2,s2,c3,s3,c4,s4;
{
  static float sin36 = 0.587785252292473;
  static float cos36 = 0.809016994374947;
  static float sin72 = 0.951056516295154;
  static float cos72 = 0.309016994374947;
  static float sin60 = 0.866025403784437;

/*
     CODING FOR FACTOR 4
*/

  ia = ja = 1;
  ib = ia+iink;
  jb = ja+jink;
  ic = ib+iink;
  jc = jb+jink;
  id = ic+iink;
  jd = jc+jink;
  for(l=1; l<=*la; l++) {
    i = ibase;
    j = jbase;

/*
  DIR$ IVDEP
*/

    for(ijk=1; ijk<=*lot; ijk++) {
      c[ja+j-1] = a[ia+i-1]+a[ic+i-1]+(a[ib+i-1]+a[id+i-1]);
      c[jc+j-1] = a[ia+i-1]+a[ic+i-1]-(a[ib+i-1]+a[id+i-1]);
      d[ja+j-1] = b[ia+i-1]+b[ic+i-1]+(b[ib+i-1]+b[id+i-1]);
      d[jc+j-1] = b[ia+i-1]+b[ic+i-1]-(b[ib+i-1]+b[id+i-1]);
      c[jb+j-1] = a[ia+i-1]-a[ic+i-1]-(b[ib+i-1]-b[id+i-1]);
      c[jd+j-1] = a[ia+i-1]-a[ic+i-1]+(b[ib+i-1]-b[id+i-1]);
      d[jb+j-1] = b[ia+i-1]-b[ic+i-1]+(a[ib+i-1]-a[id+i-1]);
      d[jd+j-1] = b[ia+i-1]-b[ic+i-1]-(a[ib+i-1]-a[id+i-1]);
      i += *inc3;
      j += *inc4;
    }
    ibase += *inc1;
    jbase += *inc2;
  }
  if(*la == m) 
    return;
  la1 = *la+1;
  jbase += jump;
  for(k=la1,D5= *la,D6=(m-k+D5)/D5; D6>0; D6--,k+=D5) {
    kb = k+k-2;
    kc = kb+kb;
    kd = kc+kb;
    c1 = trigs[kb];
    s1 = trigs[kb+1];
    c2 = trigs[kc];
    s2 = trigs[kc+1];
    c3 = trigs[kd];
    s3 = trigs[kd+1];
    for(l=1; l<=*la; l++) {
      i = ibase;
      j = jbase;

/*
  DIR$ IVDEP
*/

      for(ijk=1; ijk<=*lot; ijk++) {
	c[ja+j-1] = a[ia+i-1]+a[ic+i-1]+(a[ib+i-1]+a[id+i-1]);
	d[ja+j-1] = b[ia+i-1]+b[ic+i-1]+(b[ib+i-1]+b[id+i-1]);
	c[jc+j-1] = c2*(a[ia+i-1]+a[ic+i-1]-(a[ib+i-1]+a[id+i-1]))-s2*
	  (b[ia+i-1]+b[ic+i-1]-(b[ib+i-1]+b[id+i-1]));
	d[jc+j-1] = s2*(a[ia+i-1]+a[ic+i-1]-(a[ib+i-1]+a[id+i-1]))+c2*
	  (b[ia+i-1]+b[ic+i-1]-(b[ib+i-1]+b[id+i-1]));
	c[jb+j-1] = c1*(a[ia+i-1]-a[ic+i-1]-(b[ib+i-1]-b[id+i-1]))-s1*
	  (b[ia+i-1]-b[ic+i-1]+(a[ib+i-1]-a[id+i-1]));
	d[jb+j-1] = s1*(a[ia+i-1]-a[ic+i-1]-(b[ib+i-1]-b[id+i-1]))+c1*
	  (b[ia+i-1]-b[ic+i-1]+(a[ib+i-1]-a[id+i-1]));
	c[jd+j-1] = c3*(a[ia+i-1]-a[ic+i-1]+(b[ib+i-1]-b[id+i-1]))-s3*
	  (b[ia+i-1]-b[ic+i-1]-(a[ib+i-1]-a[id+i-1]));
	d[jd+j-1] = s3*(a[ia+i-1]-a[ic+i-1]+(b[ib+i-1]-b[id+i-1]))+c3*
	  (b[ia+i-1]-b[ic+i-1]-(a[ib+i-1]-a[id+i-1]));
	i += *inc3;
	j += *inc4;
      }
      ibase += *inc1;
      jbase += *inc2;
    }
    jbase += jump;
  }
  return;
}

void vpassm4(a, b, c, d, trigs, inc1, inc2, inc3, inc4, lot, n, ifac, la,
             m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
             D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,c1,s1,c2,s2,
             c3,s3,c4,s4)
int *inc1,*inc2,*inc3,*inc4,*lot,*n,*ifac,*la;
float a[],b[],c[],d[],trigs[];
int m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,D1,D2;
int ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8;
float c1,s1,c2,s2,c3,s3,c4,s4;
{
  static float sin36 = 0.587785252292473;
  static float cos36 = 0.809016994374947;
  static float sin72 = 0.951056516295154;
  static float cos72 = 0.309016994374947;
  static float sin60 = 0.866025403784437;
/*
     CODING FOR FACTOR 5
*/

  ia = ja = 1;
  ib = ia+iink;
  jb = ja+jink;
  ic = ib+iink;
  jc = jb+jink;
  id = ic+iink;
  jd = jc+jink;
  ie = id+iink;
  je = jd+jink;
  for(l=1; l<=*la; l++) {
    i = ibase;
    j = jbase;

/*
  DIR$ IVDEP
*/

    for(ijk=1; ijk<=*lot; ijk++) {
      c[ja+j-1] = a[ia+i-1]+(a[ib+i-1]+a[ie+i-1])+(a[ic+i-1]+a[id+i-1]);
      d[ja+j-1] = b[ia+i-1]+(b[ib+i-1]+b[ie+i-1])+(b[ic+i-1]+b[id+i-1]);
      c[jb+j-1] = a[ia+i-1]+cos72*(a[ib+i-1]+a[ie+i-1])-cos36*(a[ic+i-1]+
              a[id+i-1])-(sin72*(b[ib+i-1]-b[ie+i-1])+sin36*(b[ic+i-1]-b[id+i-1]
              ));
      c[je+j-1] = a[ia+i-1]+cos72*(a[ib+i-1]+a[ie+i-1])-cos36*(a[ic+i-1]+
              a[id+i-1])+(sin72*(b[ib+i-1]-b[ie+i-1])+sin36*(b[ic+i-1]-b[id+i-1]
              ));
      d[jb+j-1] = b[ia+i-1]+cos72*(b[ib+i-1]+b[ie+i-1])-cos36*(b[ic+i-1]+
              b[id+i-1])+(sin72*(a[ib+i-1]-a[ie+i-1])+sin36*(a[ic+i-1]-a[id+i-1]
              ));
      d[je+j-1] = b[ia+i-1]+cos72*(b[ib+i-1]+b[ie+i-1])-cos36*(b[ic+i-1]+
              b[id+i-1])-(sin72*(a[ib+i-1]-a[ie+i-1])+sin36*(a[ic+i-1]-a[id+i-1]
							     ));
      c[jc+j-1] = a[ia+i-1]-cos36*(a[ib+i-1]+a[ie+i-1])+cos72*(a[ic+i-1]+
              a[id+i-1])-(sin36*(b[ib+i-1]-b[ie+i-1])-sin72*(b[ic+i-1]-b[id+i-1]
							     ));
      c[jd+j-1] = a[ia+i-1]-cos36*(a[ib+i-1]+a[ie+i-1])+cos72*(a[ic+i-1]+
              a[id+i-1])+(sin36*(b[ib+i-1]-b[ie+i-1])-sin72*(b[ic+i-1]-b[id+i-1]
							     ));
      d[jc+j-1] = b[ia+i-1]-cos36*(b[ib+i-1]+b[ie+i-1])+cos72*(b[ic+i-1]+
              b[id+i-1])+(sin36*(a[ib+i-1]-a[ie+i-1])-sin72*(a[ic+i-1]-a[id+i-1]
							     ));
      d[jd+j-1] = b[ia+i-1]-cos36*(b[ib+i-1]+b[ie+i-1])+cos72*(b[ic+i-1]+
              b[id+i-1])-(sin36*(a[ib+i-1]-a[ie+i-1])-sin72*(a[ic+i-1]-a[id+i-1]
							     ));
      i += *inc3;
      j += *inc4;
    }
    ibase += *inc1;
    jbase += *inc2;
  }
  if(*la == m) 
    return;
  la1 = *la+1;
  jbase += jump;
  for(k=la1,D7= *la,D8=(m-k+D7)/D7; D8>0; D8--,k+=D7) {
    kb = k+k-2;
    kc = kb+kb;
    kd = kc+kb;
    ke = kd+kb;
    c1 = trigs[kb];
    s1 = trigs[kb+1];
    c2 = trigs[kc];
    s2 = trigs[kc+1];
    c3 = trigs[kd];
    s3 = trigs[kd+1];
    c4 = trigs[ke];
    s4 = trigs[ke+1];
    for(l=1; l<=*la; l++) {
      i = ibase;
      j = jbase;

/*
  DIR$ IVDEP
*/
      
      for(ijk=1; ijk<=*lot; ijk++) {
	c[ja+j-1] = a[ia+i-1]+(a[ib+i-1]+a[ie+i-1])+(a[ic+i-1]+a[id+i-1]);
	d[ja+j-1] = b[ia+i-1]+(b[ib+i-1]+b[ie+i-1])+(b[ic+i-1]+b[id+i-1]);
	c[jb+j-1] = c1*(a[ia+i-1]+cos72*(a[ib+i-1]+a[ie+i-1])-cos36*(a[
                  ic+i-1]+a[id+i-1])-(sin72*(b[ib+i-1]-b[ie+i-1])+sin36*(b[ic+i
                  -1]-b[id+i-1])))-s1*(b[ia+i-1]+cos72*(b[ib+i-1]+b[ie+i-1])-
                  cos36*(b[ic+i-1]+b[id+i-1])+(sin72*(a[ib+i-1]-a[ie+i-1])+
                  sin36*(a[ic+i-1]-a[id+i-1])));
	d[jb+j-1] = s1*(a[ia+i-1]+cos72*(a[ib+i-1]+a[ie+i-1])-cos36*(a[
                  ic+i-1]+a[id+i-1])-(sin72*(b[ib+i-1]-b[ie+i-1])+sin36*(b[ic+i
                  -1]-b[id+i-1])))+c1*(b[ia+i-1]+cos72*(b[ib+i-1]+b[ie+i-1])-
                  cos36*(b[ic+i-1]+b[id+i-1])+(sin72*(a[ib+i-1]-a[ie+i-1])+
                  sin36*(a[ic+i-1]-a[id+i-1])));
	c[je+j-1] = c4*(a[ia+i-1]+cos72*(a[ib+i-1]+a[ie+i-1])-cos36*(a[
                  ic+i-1]+a[id+i-1])+(sin72*(b[ib+i-1]-b[ie+i-1])+sin36*(b[ic+i
                  -1]-b[id+i-1])))-s4*(b[ia+i-1]+cos72*(b[ib+i-1]+b[ie+i-1])-
                  cos36*(b[ic+i-1]+b[id+i-1])-(sin72*(a[ib+i-1]-a[ie+i-1])+
                  sin36*(a[ic+i-1]-a[id+i-1])));
	d[je+j-1] = s4*(a[ia+i-1]+cos72*(a[ib+i-1]+a[ie+i-1])-cos36*(a[
                  ic+i-1]+a[id+i-1])+(sin72*(b[ib+i-1]-b[ie+i-1])+sin36*(b[ic+i
                  -1]-b[id+i-1])))+c4*(b[ia+i-1]+cos72*(b[ib+i-1]+b[ie+i-1])-
                  cos36*(b[ic+i-1]+b[id+i-1])-(sin72*(a[ib+i-1]-a[ie+i-1])+
                  sin36*(a[ic+i-1]-a[id+i-1])));
	c[jc+j-1] = c2*(a[ia+i-1]-cos36*(a[ib+i-1]+a[ie+i-1])+cos72*(a[
                  ic+i-1]+a[id+i-1])-(sin36*(b[ib+i-1]-b[ie+i-1])-sin72*(b[ic+i
                  -1]-b[id+i-1])))-s2*(b[ia+i-1]-cos36*(b[ib+i-1]+b[ie+i-1])+
                  cos72*(b[ic+i-1]+b[id+i-1])+(sin36*(a[ib+i-1]-a[ie+i-1])-
                  sin72*(a[ic+i-1]-a[id+i-1])));
	d[jc+j-1] = s2*(a[ia+i-1]-cos36*(a[ib+i-1]+a[ie+i-1])+cos72*(a[
                  ic+i-1]+a[id+i-1])-(sin36*(b[ib+i-1]-b[ie+i-1])-sin72*(b[ic+i
                  -1]-b[id+i-1])))+c2*(b[ia+i-1]-cos36*(b[ib+i-1]+b[ie+i-1])+
                  cos72*(b[ic+i-1]+b[id+i-1])+(sin36*(a[ib+i-1]-a[ie+i-1])-
                  sin72*(a[ic+i-1]-a[id+i-1])));
	c[jd+j-1] = c3*(a[ia+i-1]-cos36*(a[ib+i-1]+a[ie+i-1])+cos72*(a[
                  ic+i-1]+a[id+i-1])+(sin36*(b[ib+i-1]-b[ie+i-1])-sin72*(b[ic+i
                  -1]-b[id+i-1])))-s3*(b[ia+i-1]-cos36*(b[ib+i-1]+b[ie+i-1])+
                  cos72*(b[ic+i-1]+b[id+i-1])-(sin36*(a[ib+i-1]-a[ie+i-1])-
                  sin72*(a[ic+i-1]-a[id+i-1])));
	d[jd+j-1] = s3*(a[ia+i-1]-cos36*(a[ib+i-1]+a[ie+i-1])+cos72*(a[
                  ic+i-1]+a[id+i-1])+(sin36*(b[ib+i-1]-b[ie+i-1])-sin72*(b[ic+i
                  -1]-b[id+i-1])))+c3*(b[ia+i-1]-cos36*(b[ib+i-1]+b[ie+i-1])+
                  cos72*(b[ic+i-1]+b[id+i-1])-(sin36*(a[ib+i-1]-a[ie+i-1])-
                  sin72*(a[ic+i-1]-a[id+i-1])));
	i += *inc3;
	j += *inc4;
      }
      ibase += *inc1;
      jbase += *inc2;
    }
    jbase += jump;
  }
  return;
}


/**************************************************************************
* Function: vpassm()
*
***************************************************************************/
void vpassm(a, b, c, d, trigs, inc1, inc2, inc3, inc4, lot, n, ifac, la)
/*
     SUBROUTINE ^VPASSM^ - MULTIPLE VERSION OF ^VPASSA^
     PERFORMS ONE PASS THROUGH DATA
     AS PART OF MULTIPLE COMPLEX (INVERSE) FFT ROUTINE
     A IS FIRST REAL INPUT VECTOR
     B IS FIRST IMAGINARY INPUT VECTOR
     C IS FIRST REAL OUTPUT VECTOR
     D IS FIRST IMAGINARY OUTPUT VECTOR
     TRIGS IS PRECALCULATED TABLE OF SINES \ COSINES
     INC1 IS ADDRESSING INCREMENT FOR A AND B
     INC2 IS ADDRESSING INCREMENT FOR C AND D
     INC3 IS ADDRESSING INCREMENT BETWEEN A^S \ B^S
     INC4 IS ADDRESSING INCREMENT BETWEEN C^S \ D^S
     LOT IS THE NUMBER OF VECTORS
     N IS LENGTH OF VECTORS
     IFAC IS CURRENT FACTOR OF N
     LA IS PRODUCT OF PREVIOUS FACTORS
*/

int *inc1,*inc2,*inc3,*inc4,*lot,*n,*ifac,*la;
float a[],b[],c[],d[],trigs[];
{
  static float sin36 = 0.587785252292473;
  static float cos36 = 0.809016994374947;
  static float sin72 = 0.951056516295154;
  static float cos72 = 0.309016994374947;
  static float sin60 = 0.866025403784437;
  static int m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,D1,D2,
    ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8;
  static float c1,s1,c2,s2,c3,s3,c4,s4;

  m = *n/ *ifac;
  iink = m**inc1;
  jink = *la**inc2;
  jump = (*ifac-1)*jink;
  ibase = jbase = 0;
  igo = *ifac-1;
  if(igo > 4) return;
  switch(igo) {
    case 1:
      vpassm1(a,b,c,d,trigs,inc1,inc2,inc3,inc4,lot,n,ifac,la,
              m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
              D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,
              c1,s1,c2,s2,c3,s3,c4,s4);
      break;
    case 2: 
      vpassm2(a,b,c,d,trigs,inc1,inc2,inc3,inc4,lot,n,ifac,la,
              m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
              D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,
              c1,s1,c2,s2,c3,s3,c4,s4);
      break;
    case 3: 
      vpassm3(a,b,c,d,trigs,inc1,inc2,inc3,inc4,lot,n,ifac,la,
              m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
              D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,
              c1,s1,c2,s2,c3,s3,c4,s4);
      break;
    case 4:
      vpassm4(a,b,c,d,trigs,inc1,inc2,inc3,inc4,lot,n,ifac,la,
              m,iink,jink,jump,ibase,jbase,igo,ia,ja,ib,jb,l,i,j,ijk,la1,k,kb,
              D1,D2,ic,jc,kc,D3,D4,id,jd,kd,D5,D6,ie,je,ke,D7,D8,
              c1,s1,c2,s2,c3,s3,c4,s4);
      break;
    default: break;
  }
}


/***************************************************************************
* Function: fft99()
*
*****************************************************************************/
void fft99(a, work, trigs, ifax, inc, jump, n, lot, isign)
/*
 PURPOSE      PERFORMS MULTIPLE FAST FOURIER TRANSFORMS.  THIS PACKAGE
              WILL PERFORM A NUMBER OF SIMULTANEOUS REAL/HALF-COMPLEX
              PERIODIC FOURIER TRANSFORMS OR CORRESPONDING INVERSE
              TRANSFORMS, I.E.  GIVEN A SET OF REAL DATA VECTORS, THE
              PACKAGE RETURNS A SET OF #HALF-COMPLEX# FOURIER
              COEFFICIENT VECTORS, OR VICE VERSA.  THE LENGTH OF THE
              TRANSFORMS MUST BE AN EVEN NUMBER GREATER THAN 4 THAT HAS
              NO OTHER FACTORS EXCEPT POSSIBLY POWERS OF 2, 3, AND 5.
              THIS IS AN ALL FORTRAN VERSION OF THE CRAYLIB PACKAGE
              THAT IS MOSTLY WRITTEN IN CAL.
              THE PACKAGE FFT99F CONTAINS SEVERAL USER-LEVEL ROUTINES:
            SUBROUTINE FFTFAX
                AN INITIALIZATION ROUTINE THAT MUST BE CALLED ONCE
                BEFORE A SEQUENCE OF CALLS TO THE FFT ROUTINES
                (PROVIDED THAT N IS NOT CHANGED).
            SUBROUTINES FFT99 AND FFT991
                TWO FFT ROUTINES THAT RETURN SLIGHTLY DIFFERENT
                ARRANGEMENTS OF THE DATA IN GRIDPOINT SPACE.
 ACCESS       THIS FORTRAN VERSION MAY BE ACCESSED WITH
                   *FORTRAN,P=XLIB,SN=FFT99F
              TO ACCESS THE CRAY OBJECT CODE, CALLING THE USER ENTRY
              POINTS FROM A CRAY PROGRAM IS SUFFICIENT.  THE SOURCE
              FORTRAN AND CAL CODE FOR THE CRAYLIB VERSION MAY BE
              ACCESSED USING
                   FETCH P=CRAYLIB,SN=FFT99
                   FETCH P=CRAYLIB,SN=CAL99
 USAGE        LET N BE OF THE FORM 2**P * 3**Q * 5**R, WHERE P .GE. 1,
              Q .GE. 0, AND R .GE. 0.  THEN A TYPICAL SEQUENCE OF
              CALLS TO TRANSFORM A GIVEN SET OF REAL VECTORS OF LENGTH
              N TO A SET OF #HALF-COMPLEX# FOURIER COEFFICIENT VECTORS
              OF LENGTH N IS
                   DIMENSION IFAX(13),TRIGS(3*N/2+1),A(M*(N+2)),
                  +          WORK(M*(N+1))
                   CALL FFTFAX (N, IFAX, TRIGS)
                   CALL FFT99 (A,WORK,TRIGS,IFAX,INC,JUMP,N,M,ISIGN)
              SEE THE INDIVIDUAL WRITE-UPS FOR FFTFAX, FFT99, AND
              FFT991 BELOW, FOR A DETAILED DESCRIPTION OF THE
              ARGUMENTS.
 HISTORY      THE PACKAGE WAS WRITTEN BY CLIVE TEMPERTON AT ECMWF IN
              NOVEMBER, 1978.  IT WAS MODIFIED, DOCUMENTED, AND TESTED
              FOR NCAR BY RUSS REW IN SEPTEMBER, 1980.
-----------------------------------------------------------------------
 SUBROUTINE FFTFAX (N,IFAX,TRIGS)
 PURPOSE      A SET-UP ROUTINE FOR FFT99 AND FFT991.  IT NEED ONLY BE
              CALLED ONCE BEFORE A SEQUENCE OF CALLS TO THE FFT
              ROUTINES (PROVIDED THAT N IS NOT CHANGED).
 ARGUMENT     IFAX(13),TRIGS(3*N/2+1)
 DIMENSIONS
 ARGUMENTS
 ON INPUT     N
               AN EVEN NUMBER GREATER THAN 4 THAT HAS NO PRIME FACTOR
               GREATER THAN 5.  N IS THE LENGTH OF THE TRANSFORMS (SEE
               THE DOCUMENTATION FOR FFT99 AND FFT991 FOR THE
               DEFINITIONS OF THE TRANSFORMS).
              IFAX
               AN INTEGER ARRAY.  THE NUMBER OF ELEMENTS ACTUALLY USED
               WILL DEPEND ON THE FACTORIZATION OF N.  DIMENSIONING
               IFAX FOR 13 SUFFICES FOR ALL N LESS THAN A MILLION.
              TRIGS
               A FLOATING POINT ARRAY OF DIMENSION 3*N/2 IF N/2 IS
               EVEN, OR 3*N/2+1 IF N/2 IS ODD.
 ON OUTPUT    IFAX
               CONTAINS THE FACTORIZATION OF N/2.  IFAX(1) IS THE
               NUMBER OF FACTORS, AND THE FACTORS THEMSELVES ARE STORED
               IN IFAX(2),IFAX(3),...  IF FFTFAX IS CALLED WITH N ODD,
               OR IF N HAS ANY PRIME FACTORS GREATER THAN 5, IFAX(1)
               IS SET TO -99.
              TRIGS
               AN ARRAY OF TRIGNOMENTRIC FUNCTION VALUES SUBSEQUENTLY
               USED BY THE FFT ROUTINES.
-----------------------------------------------------------------------
 SUBROUTINE FFT991 (A,WORK,TRIGS,IFAX,INC,JUMP,N,M,ISIGN)
                       AND
 SUBROUTINE FFT99 (A,WORK,TRIGS,IFAX,INC,JUMP,N,M,ISIGN)
 PURPOSE      PERFORM A NUMBER OF SIMULTANEOUS REAL/HALF-COMPLEX
              PERIODIC FOURIER TRANSFORMS OR CORRESPONDING INVERSE
              TRANSFORMS, USING ORDINARY SPATIAL ORDER OF GRIDPOINT
              VALUES (FFT991) OR EXPLICIT CYCLIC CONTINUITY IN THE
              GRIDPOINT VALUES (FFT99).  GIVEN A SET
              OF REAL DATA VECTORS, THE PACKAGE RETURNS A SET OF
              #HALF-COMPLEX# FOURIER COEFFICIENT VECTORS, OR VICE
              VERSA.  THE LENGTH OF THE TRANSFORMS MUST BE AN EVEN
              NUMBER THAT HAS NO OTHER FACTORS EXCEPT POSSIBLY POWERS
              OF 2, 3, AND 5.  THESE VERSION OF FFT991 AND FFT99 ARE
              OPTIMIZED FOR USE ON THE CRAY-1.
 ARGUMENT     A(M*(N+2)), WORK(M*(N+1)), TRIGS(3*N/2+1), IFAX(13)
 DIMENSIONS
 ARGUMENTS
 ON INPUT     A
               AN ARRAY OF LENGTH M*(N+2) CONTAINING THE INPUT DATA
               OR COEFFICIENT VECTORS.  THIS ARRAY IS OVERWRITTEN BY
               THE RESULTS.
              WORK
               A WORK ARRAY OF DIMENSION M*(N+1)
              TRIGS
               AN ARRAY SET UP BY FFTFAX, WHICH MUST BE CALLED FIRST.
              IFAX
               AN ARRAY SET UP BY FFTFAX, WHICH MUST BE CALLED FIRST.
              INC
               THE INCREMENT (IN WORDS) BETWEEN SUCCESSIVE ELEMENTS OF
               EACH DATA OR COEFFICIENT VECTOR (E.G.  INC=1 FOR
               CONSECUTIVELY STORED DATA).
              JUMP
               THE INCREMENT (IN WORDS) BETWEEN THE FIRST ELEMENTS OF
               SUCCESSIVE DATA OR COEFFICIENT VECTORS.  ON THE CRAY-1,
               TRY TO ARRANGE DATA SO THAT JUMP IS NOT A MULTIPLE OF 8
               (TO AVOID MEMORY BANK CONFLICTS).  FOR CLARIFICATION OF
               INC AND JUMP, SEE THE EXAMPLES BELOW.
              N
               THE LENGTH OF EACH TRANSFORM (SEE DEFINITION OF
               TRANSFORMS, BELOW).
              M
               THE NUMBER OF TRANSFORMS TO BE DONE SIMULTANEOUSLY.
              ISIGN
               = +1 FOR A TRANSFORM FROM FOURIER COEFFICIENTS TO
                    GRIDPOINT VALUES.
               = -1 FOR A TRANSFORM FROM GRIDPOINT VALUES TO FOURIER
                    COEFFICIENTS.
 ON OUTPUT    A
               IF ISIGN = +1, AND M COEFFICIENT VECTORS ARE SUPPLIED
               EACH CONTAINING THE SEQUENCE:
               A(0),B(0),A(1),B(1),...,A(N/2),B(N/2)  (N+2 VALUES)
               THEN THE RESULT CONSISTS OF M DATA VECTORS EACH
               CONTAINING THE CORRESPONDING N+2 GRIDPOINT VALUES:
               FOR FFT991, X(0), X(1), X(2),...,X(N-1),0,0.
               FOR FFT99, X(N-1),X(0),X(1),X(2),...,X(N-1),X(0).
                   (EXPLICIT CYCLIC CONTINUITY)
               WHEN ISIGN = +1, THE TRANSFORM IS DEFINED BY:
                 X(J)=SUM(K=0,...,N-1)(C(K)*EXP(2*I*J*K*PI/N))
                 WHERE C(K)=A(K)+I*B(K) AND C(N-K)=A(K)-I*B(K)
                 AND I=SQRT (-1)
               IF ISIGN = -1, AND M DATA VECTORS ARE SUPPLIED EACH
               CONTAINING A SEQUENCE OF GRIDPOINT VALUES X(J) AS
               DEFINED ABOVE, THEN THE RESULT CONSISTS OF M VECTORS
               EACH CONTAINING THE CORRESPONDING FOURIER COFFICIENTS
               A(K), B(K), 0 .LE. K .LE N/2.
               WHEN ISIGN = -1, THE INVERSE TRANSFORM IS DEFINED BY:
                 C(K)=(1/N)*SUM(J=0,...,N-1)(X(J)*EXP(-2*I*J*K*PI/N))
                 WHERE C(K)=A(K)+I*B(K) AND I=SQRT(-1)
               A CALL WITH ISIGN=+1 FOLLOWED BY A CALL WITH ISIGN=-1
               (OR VICE VERSA) RETURNS THE ORIGINAL DATA.
               NOTE: THE FACT THAT THE GRIDPOINT VALUES X(J) ARE REAL
               IMPLIES THAT B(0)=B(N/2)=0.  FOR A CALL WITH ISIGN=+1,
               IT IS NOT ACTUALLY NECESSARY TO SUPPLY THESE ZEROS.
 EXAMPLES      GIVEN 19 DATA VECTORS EACH OF LENGTH 64 (+2 FOR EXPLICIT
               CYCLIC CONTINUITY), COMPUTE THE CORRESPONDING VECTORS OF
               FOURIER COEFFICIENTS.  THE DATA MAY, FOR EXAMPLE, BE
               ARRANGED LIKE THIS:
 FIRST DATA   A(1)=    . . .                A(66)=             A(70)
 VECTOR       X(63) X(0) X(1) X(2) ... X(63) X(0)  (4 EMPTY LOCATIONS)
 SECOND DATA  A(71)=   . . .                                  A(140)
 VECTOR       X(63) X(0) X(1) X(2) ... X(63) X(0)  (4 EMPTY LOCATIONS)
               AND SO ON.  HERE INC=1, JUMP=70, N=64, M=19, ISIGN=-1,
               AND FFT99 SHOULD BE USED (BECAUSE OF THE EXPLICIT CYCLIC
               CONTINUITY).
               ALTERNATIVELY THE DATA MAY BE ARRANGED LIKE THIS:
                FIRST         SECOND                          LAST
                DATA          DATA                            DATA
                VECTOR        VECTOR                          VECTOR
                 A(1)=         A(2)=                           A(19)=
                 X(63)         X(63)       . . .               X(63)
        A(20)=   X(0)          X(0)        . . .               X(0)
        A(39)=   X(1)          X(1)        . . .               X(1)
                  .             .                               .
                  .             .                               .
                  .             .                               .
               IN WHICH CASE WE HAVE INC=19, JUMP=1, AND THE REMAINING
               PARAMETERS ARE THE SAME AS BEFORE.  IN EITHER CASE, EACH
               COEFFICIENT VECTOR OVERWRITES THE CORRESPONDING INPUT
               DATA VECTOR.
-----------------------------------------------------------------------
*/

int ifax[],*inc,*jump,*n,*lot,*isign;
float a[],work[],trigs[];
{
  extern void fft99a(),vpassm(),fft99b();
  static int K1 = 2;
  static int nfax,nx,nh,ink,igo,ibase,jbase,l,i,j,m,ia,la,k,ib;
/*
     SUBROUTINE ^FFT99^ - MULTIPLE FAST REAL PERIODIC TRANSFORM
     CORRESPONDING TO OLD SCALAR ROUTINE FFT9
     PROCEDURE USED TO CONVERT TO HALF-LENGTH COMPLEX TRANSFORM
     IS GIVEN BY COOLEY, LEWIS AND WELCH (J. SOUND VIB., VOL. 12
     (1970), 315-337)
     A IS THE ARRAY CONTAINING INPUT AND OUTPUT DATA
     WORK IS AN AREA OF SIZE (N+1)*LOT
     TRIGS IS A PREVIOUSLY PREPARED LIST OF TRIG FUNCTION VALUES
     IFAX IS A PREVIOUSLY PREPARED LIST OF FACTORS OF N/2
     INC IS THE INCREMENT WITHIN EACH DATA #VECTOR#
         (E.G. INC=1 FOR CONSECUTIVELY STORED DATA)
     JUMP IS THE INCREMENT BETWEEN THE START OF EACH DATA VECTOR
     N IS THE LENGTH OF THE DATA VECTORS
     LOT IS THE NUMBER OF DATA VECTORS
     ISIGN = +1 FOR TRANSFORM FROM SPECTRAL TO GRIDPOINT
           = -1 FOR TRANSFORM FROM GRIDPOINT TO SPECTRAL
     ORDERING OF COEFFICIENTS:
         A(0),B(0),A(1),B(1),A(2),B(2),...,A(N/2),B(N/2)
         WHERE B(0)=B(N/2)=0; (N+2) LOCATIONS REQUIRED
     ORDERING OF DATA:
         X(N-1),X(0),X(1),X(2),...,X(N),X(0)
         I.E. EXPLICIT CYCLIC CONTINUITY; (N+2) LOCATIONS REQUIRED
     VECTORIZATION IS ACHIEVED ON CRAY BY DOING THE TRANSFORMS IN
     PARALLEL
     *** N.B. N IS ASSUMED TO BE AN EVEN NUMBER
     DEFINITION OF TRANSFORMS:
     -------------------------
     ISIGN=+1: X(J)=SUM(K=0,...,N-1)(C(K)*EXP(2*I*J*K*PI/N))
         WHERE C(K)=A(K)+I*B(K) AND C(N-K)=A(K)-I*B(K)
     ISIGN=-1: A(K)=(1/N)*SUM(J=0,...,N-1)(X(J)*COS(2*J*K*PI/N))
               B(K)=-(1/N)*SUM(J=0,...,N-1)(X(J)*SIN(2*J*K*PI/N))
 THE FOLLOWING CALL IS FOR MONITORING LIBRARY USE AT NCAR
      CALL Q8QST4 ( 4HXLIB, 6HFFT99F, 5HFFT99, 10HVERSION 01)
*/

  nfax = ifax[0];
  nx = *n+1;
  nh = *n/2;
  ink = *inc+*inc;
  if(*isign == 1) goto S30;

/*
     IF NECESSARY, TRANSFER DATA TO WORK AREA
*/

  igo = 50;
  if(fifi2mod(nfax,2) == 1) goto S40;
  ibase = *inc+1;
  jbase = 1;
  for(l=1; l<=*lot; l++) {
    i = ibase;
    j = jbase;

/*
  DIR$ IVDEP
*/

    for(m=1; m<=*n; m++) {
      work[j-1] = a[i-1];
      i += *inc;
      j += 1;
    }
    ibase += *jump;
    jbase += nx;
  }
  igo = 60;
  goto S40;
S30:

/*
     PREPROCESSING (ISIGN=+1)
     ------------------------
*/

  fft99a(a,work,trigs,inc,jump,n,lot);
  igo = 60;
S40:

/*
     COMPLEX TRANSFORM
     -----------------
*/

  ia = *inc+1;
  la = 1;
  for(k=1; k<=nfax; k++) {
    if(igo == 60) goto S60;
    vpassm(&a[ia-1],&a[ia+*inc-1],&work[0],&work[1],trigs,&ink,&K1,jump,&nx,
	   lot,&nh,&ifax[k],&la);
    igo = 60;
    goto S70;
  S60:
    vpassm(&work[0],&work[1],&a[ia-1],&a[ia+*inc-1],trigs,&K1,&ink,&nx,jump,
	   lot,&nh,&ifax[k],&la);
    igo = 50;
  S70:
    la *= ifax[k];
  }
  if(*isign == -1) goto S130;

/*
     IF NECESSARY, TRANSFER DATA FROM WORK AREA
*/

  if(fifi2mod(nfax,2) == 1) goto S110;
  ibase = 1;
  jbase = ia;
  for(l=1; l<=*lot; l++) {
    i = ibase;
    j = jbase;

/*
  DIR$ IVDEP
*/

    for(m=1; m<=*n; m++) {
      a[j-1] = work[i-1];
      i += 1;
      j += *inc;
    }
    ibase += nx;
    jbase += *jump;
  }
S110:

/*
     FILL IN CYCLIC BOUNDARY POINTS
*/

  ia = 1;
  ib = *n**inc+1;
/*
  DIR$ IVDEP
*/

  for(l=1; l<=*lot; l++) {
    a[ia-1] = a[ib-1];
    a[ib+*inc-1] = a[ia+*inc-1];
    ia += *jump;
    ib += *jump;
  }
  goto S140;
S130:

/*
     POSTPROCESSING (ISIGN=-1):
     --------------------------
*/

  fft99b(work,a,trigs,inc,jump,n,lot);
S140:
  return;
}


/*****************************************************************************
* Function: fft99a()
*
*******************************************************************************/
void fft99a(a, work, trigs, inc, jump, n, lot)
int *inc,*jump,*n,*lot;
float a[],work[],trigs[];
{
  static int nh,nx,ink,ia,ib,ja,jb,l,iabase,ibbase,jabase,jbbase,k;
  static float c,s;

/*
     SUBROUTINE FFT99A - PREPROCESSING STEP FOR FFT99, ISIGN=+1
     (SPECTRAL TO GRIDPOINT TRANSFORM)
*/

  nh = *n/2;
  nx = *n+1;
  ink = *inc+*inc;

/*
     A(0) AND A(N/2)
*/

  ia = 1;
  ib = *n**inc+1;
  ja = 1;
  jb = 2;

/*
  DIR$ IVDEP
*/

  for(l=1; l<=*lot; l++) {
    work[ja-1] = a[ia-1]+a[ib-1];
    work[jb-1] = a[ia-1]-a[ib-1];
    ia += *jump;
    ib += *jump;
    ja += nx;
    jb += nx;
  }

/*
     REMAINING WAVENUMBERS
*/

  iabase = 2**inc+1;
  ibbase = (*n-2)**inc+1;
  jabase = 3;
  jbbase = *n-1;
  for(k=3; k<=nh; k+=2) {
    ia = iabase;
    ib = ibbase;
    ja = jabase;
    jb = jbbase;
    c = trigs[*n+k-1];
    s = trigs[*n+k];

/*
  DIR$ IVDEP
*/

    for(l=1; l<=*lot; l++) {
      work[ja-1] = a[ia-1]+a[ib-1]-(s*(a[ia-1]-a[ib-1])+c*(a[ia+*inc-1]+a[
									  ib+*inc-1]));
      work[jb-1] = a[ia-1]+a[ib-1]+(s*(a[ia-1]-a[ib-1])+c*(a[ia+*inc-1]+a[
									  ib+*inc-1]));
      work[ja] = c*(a[ia-1]-a[ib-1])-s*(a[ia+*inc-1]+a[ib+*inc-1])+(a[ia+*
              inc-1]-a[ib+*inc-1]);
      work[jb] = c*(a[ia-1]-a[ib-1])-s*(a[ia+*inc-1]+a[ib+*inc-1])-(a[ia+*
              inc-1]-a[ib+*inc-1]);
      ia += *jump;
      ib += *jump;
      ja += nx;
      jb += nx;
    }
    iabase += ink;
    ibbase -= ink;
    jabase += 2;
    jbbase -= 2;
  }
  if(iabase != ibbase) 
    goto S50;

/*
     WAVENUMBER N/4 (IF IT EXISTS)
*/

  ia = iabase;
  ja = jabase;

/*
  DIR$ IVDEP
*/
  
  for(l=1; l<=*lot; l++) {
    work[ja-1] = 2.0*a[ia-1];
    work[ja] = -2.0*a[ia+*inc-1];
    ia += *jump;
    ja += nx;
  }
S50:
  return;
}


/*******************************************************************************
* Function: fft99b()
*
********************************************************************************/
void fft99b(work, a, trigs, inc, jump, n, lot)
int *inc,*jump,*n,*lot;
float work[],a[],trigs[];
{
  static int nh,nx,ink,ia,ib,ja,jb,l,iabase,ibbase,jabase,jbbase,k;
  static float scale,c,s;

/*
     SUBROUTINE FFT99B - POSTPROCESSING STEP FOR FFT99, ISIGN=-1
     (GRIDPOINT TO SPECTRAL TRANSFORM)
*/

  nh = *n/2;
  nx = *n+1;
  ink = *inc+*inc;

/*
     A(0) AND A(N/2)
*/

  scale = 1.0/(float)*n;
  ia = 1;
  ib = 2;
  ja = 1;
  jb = *n**inc+1;

/*
  DIR$ IVDEP
*/

  for(l=1; l<=*lot; l++) {
    a[ja-1] = scale*(work[ia-1]+work[ib-1]);
    a[jb-1] = scale*(work[ia-1]-work[ib-1]);
    a[ja+*inc-1] = a[jb+*inc-1] = 0.0;
    ia += nx;
    ib += nx;
    ja += *jump;
    jb += *jump;
  }

/*
     REMAINING WAVENUMBERS
*/

  scale = 0.5*scale;
  iabase = 3;
  ibbase = *n-1;
  jabase = 2**inc+1;
  jbbase = (*n-2)**inc+1;
  for(k=3; k<=nh; k+=2) {
    ia = iabase;
    ib = ibbase;
    ja = jabase;
    jb = jbbase;
    c = trigs[*n+k-1];
    s = trigs[*n+k];

/*
  DIR$ IVDEP
*/

    for(l=1; l<=*lot; l++) {
      a[ja-1] = scale*(work[ia-1]+work[ib-1]+(c*(work[ia]+work[ib])+s*
					      (work[ia-1]-work[ib-1])));
      a[jb-1] = scale*(work[ia-1]+work[ib-1]-(c*(work[ia]+work[ib])+s*
					      (work[ia-1]-work[ib-1])));
      a[ja+*inc-1] = scale*(c*(work[ia-1]-work[ib-1])-s*(work[ia]+work[ib]
							 )+(work[ib]-work[ia]));
      a[jb+*inc-1] = scale*(c*(work[ia-1]-work[ib-1])-s*(work[ia]+work[ib]
							 )-(work[ib]-work[ia]));
      ia += nx;
      ib += nx;
      ja += *jump;
      jb += *jump;
    }
    iabase += 2;
    ibbase -= 2;
    jabase += ink;
    jbbase -= ink;
  }
  if(iabase != ibbase) 
    goto S50;

/*
     WAVENUMBER N/4 (IF IT EXISTS)
*/

  ia = iabase;
  ja = jabase;
  scale = 2.0*scale;

/*
  DIR$ IVDEP
*/

  for(l=1; l<=*lot; l++) {
    a[ja-1] = scale*work[ia-1];
    a[ja+*inc-1] = -(scale*work[ia]);
    ia += nx;
    ja += *jump;
  }
S50:
  return;
}


/******************************************************************************
* Function: fft991()
*
********************************************************************************/
void fft991(a, work, trigs, ifax, inc, jump, n, lot, isign)
int ifax[],*inc,*jump,*n,*lot,*isign;
float a[],work[],trigs[];
{
  extern void fft99a(),vpassm(),fft99b();
  static int K1 = 2;
  static int nfax,nx,nh,ink,igo,ibase,jbase,l,i,j,m,ia,la,k,ib;

/*
     SUBROUTINE ^FFT991^ - MULTIPLE REAL/HALF-COMPLEX PERIODIC
     FAST FOURIER TRANSFORM
     SAME AS FFT99 EXCEPT THAT ORDERING OF DATA CORRESPONDS TO
     THAT IN MRFFT2
     PROCEDURE USED TO CONVERT TO HALF-LENGTH COMPLEX TRANSFORM
     IS GIVEN BY COOLEY, LEWIS AND WELCH (J. SOUND VIB., VOL. 12
     (1970), 315-337)
     A IS THE ARRAY CONTAINING INPUT AND OUTPUT DATA
     WORK IS AN AREA OF SIZE (N+1)*LOT
     TRIGS IS A PREVIOUSLY PREPARED LIST OF TRIG FUNCTION VALUES
     IFAX IS A PREVIOUSLY PREPARED LIST OF FACTORS OF N/2
     INC IS THE INCREMENT WITHIN EACH DATA #VECTOR#
         (E.G. INC=1 FOR CONSECUTIVELY STORED DATA)
     JUMP IS THE INCREMENT BETWEEN THE START OF EACH DATA VECTOR
     N IS THE LENGTH OF THE DATA VECTORS
     LOT IS THE NUMBER OF DATA VECTORS
     ISIGN = +1 FOR TRANSFORM FROM SPECTRAL TO GRIDPOINT
           = -1 FOR TRANSFORM FROM GRIDPOINT TO SPECTRAL
     ORDERING OF COEFFICIENTS:
         A(0),B(0),A(1),B(1),A(2),B(2),...,A(N/2),B(N/2)
         WHERE B(0)=B(N/2)=0; (N+2) LOCATIONS REQUIRED
     ORDERING OF DATA:
         X(0),X(1),X(2),...,X(N-1)
     VECTORIZATION IS ACHIEVED ON CRAY BY DOING THE TRANSFORMS IN
     PARALLEL
     *** N.B. N IS ASSUMED TO BE AN EVEN NUMBER
     DEFINITION OF TRANSFORMS:
     -------------------------
     ISIGN=+1: X(J)=SUM(K=0,...,N-1)(C(K)*EXP(2*I*J*K*PI/N))
         WHERE C(K)=A(K)+I*B(K) AND C(N-K)=A(K)-I*B(K)
     ISIGN=-1: A(K)=(1/N)*SUM(J=0,...,N-1)(X(J)*COS(2*J*K*PI/N))
               B(K)=-(1/N)*SUM(J=0,...,N-1)(X(J)*SIN(2*J*K*PI/N))
 THE FOLLOWING CALL IS FOR MONITORING LIBRARY USE AT NCAR
      CALL Q8QST4 ( 4HXLIB, 6HFFT99F, 6HFFT991, 10HVERSION 01)
*/

  nfax = ifax[0];
  nx = *n+1;
  nh = *n/2;
  ink = *inc+*inc;
  if(*isign == 1) goto S30;

/*
     IF NECESSARY, TRANSFER DATA TO WORK AREA
*/

  igo = 50;
  if(fifi2mod(nfax,2) == 1) goto S40;
  ibase = jbase = 1;
  for(l=1; l<=*lot; l++) {
    i = ibase;
    j = jbase;

/*
  DIR$ IVDEP
*/

    for(m=1; m<=*n; m++) {
      work[j-1] = a[i-1];
      i += *inc;
      j += 1;
    }
    ibase += *jump;
    jbase += nx;
  }
  igo = 60;
  goto S40;
S30:

/*
     PREPROCESSING (ISIGN=+1)
     ------------------------
*/

  fft99a(a,work,trigs,inc,jump,n,lot);
  igo = 60;
S40:

/*
     COMPLEX TRANSFORM
     -----------------
*/

  ia = la = 1;
  for(k=1; k<=nfax; k++) {
    if(igo == 60) goto S60;
    vpassm(&a[ia-1],&a[ia+*inc-1],&work[0],&work[1],trigs,&ink,&K1,jump,&nx,
	   lot,&nh,&ifax[k],&la);
    igo = 60;
    goto S70;
  S60:
    vpassm(&work[0],&work[1],&a[ia-1],&a[ia+*inc-1],trigs,&K1,&ink,&nx,jump,
	   lot,&nh,&ifax[k],&la);
    igo = 50;
  S70:
    la *= ifax[k];
  }
  if(*isign == -1) goto S130;

/*
     IF NECESSARY, TRANSFER DATA FROM WORK AREA
*/

  if(fifi2mod(nfax,2) == 1) goto S110;
  ibase = jbase = 1;
  for(l=1; l<=*lot; l++) {
    i = ibase;
    j = jbase;
/*
  DIR$ IVDEP
*/

    for(m=1; m<=*n; m++) {
      a[j-1] = work[i-1];
      i += 1;
      j += *inc;
    }
    ibase += nx;
    jbase += *jump;
  }
S110:

/*
     FILL IN ZEROS AT END
*/

  ib = *n**inc+1;

/*
  DIR$ IVDEP
*/

  for(l=1; l<=*lot; l++) {
    a[ib-1] = a[ib+*inc-1] = 0.0;
    ib += *jump;
  }
  goto S140;
S130:

/*
     POSTPROCESSING (ISIGN=-1):
     --------------------------
*/

  fft99b(work,a,trigs,inc,jump,n,lot);
S140:
  return;
}


/*************************************************************************
* Function: fftfax()
*
***************************************************************************/
void fftfax(n, ifax, trigs)
/*
 MODE 3 IS USED FOR REAL/HALF-COMPLEX TRANSFORMS.  IT IS POSSIBLE
 TO DO COMPLEX/COMPLEX TRANSFORMS WITH OTHER VALUES OF MODE, BUT
 DOCUMENTATION OF THE DETAILS WERE NOT AVAILABLE WHEN THIS ROUTINE
 WAS WRITTEN.
*/

int *n,ifax[];
float trigs[];
{
  extern void fax(),fftrig();
  static int mode = 3;
  static int i;
  
  fax(ifax,n,&mode);
  i = ifax[0];
  if(ifax[i] > 5 || *n <= 4) ifax[0] = -99;
  if(ifax[0] <= 0)
    puts(" FFTFAX -- INVALID N");

  fftrig(trigs,n,&mode);
  return;
}


/****************************************************************************
* Function: fax()
*
*****************************************************************************/
void fax(ifax, n, mode)
int ifax[],*n,*mode;
{
  static int nn,k,l,inc,nfax,ii,istop,i,item;

  nn = *n;
  if(abs(*mode) == 1) 
    goto S10;
  if(abs(*mode) == 8) 
    goto S10;
  nn = *n/2;
  if(nn+nn == *n) 
    goto S10;
  ifax[0] = -99;
  return;
S10:
  k = 1;
S20:

/*
     TEST FOR FACTORS OF 4
*/

  if(fifi2mod(nn,4) != 0) 
    goto S30;
  k += 1;
  ifax[k-1] = 4;
  nn /= 4;
  if(nn == 1) 
    goto S80;
  goto S20;
S30:

/*
     TEST FOR EXTRA FACTOR OF 2
*/

  if(fifi2mod(nn,2) != 0) goto S40;
  k += 1;
  ifax[k-1] = 2;
  nn /= 2;
  if(nn == 1) goto S80;
S40:

/*
     TEST FOR FACTORS OF 3
*/

  if(fifi2mod(nn,3) != 0) goto S50;
  k += 1;
  ifax[k-1] = 3;
  nn /= 3;
  if(nn == 1) 
    goto S80;
  goto S40;
S50:

/*
     NOW FIND REMAINING FACTORS
*/
  
  l = 5;
  inc = 2;
S60:

/*
     INC ALTERNATELY TAKES ON VALUES 2 AND 4
*/

  if(fifi2mod(nn,l) != 0) 
    goto S70;
  k += 1;
  ifax[k-1] = l;
  nn /= l;
  if(nn == 1) 
    goto S80;
  goto S60;
S70:
  l += inc;
  inc = 6-inc;
  goto S60;
S80:
  ifax[0] = k-1;

/*
     IFAX(1) CONTAINS NUMBER OF FACTORS
*/
    nfax = ifax[0];

/*
     SORT FACTORS INTO ASCENDING ORDER
*/
    if(nfax == 1) 
      goto S110;
    for(ii=2; ii<=nfax; ii++) {
      istop = nfax+2-ii;
      for(i=2; i<=istop; i++) {
	if(ifax[i] >= ifax[i-1]) 
	  goto S90;
	item = ifax[i-1];
	ifax[i-1] = ifax[i];
	ifax[i] = item;
      S90:
	continue;
      }
    }
S110:
    return;
}


/*****************************************************************************
* Function: fftrig()
*
*******************************************************************************/
void fftrig(trigs, n, mode)
int *n,*mode;
float trigs[];
{
  static int imode,nn,l,i,nh,la;
  static float pi,del,angle;

  pi = 2.0*asin(1.0);
  imode = abs(*mode);
  nn = *n;
  if(imode > 1 && imode < 6) 
    nn = *n/2;
  del = (pi+pi)/(float)nn;
  l = nn+nn;
  for(i=1; i<=l; i+=2) {
    angle = 0.5*(float)(i-1)*del;
    trigs[i-1] = cos(angle);
    trigs[i] = sin(angle);
  }
  if(imode == 1) 
    return;
  if(imode == 8) 
    return;
  del = 0.5*del;
  nh = (nn+1)/2;
  l = nh+nh;
  la = nn+nn;
  for(i=1; i<=l; i+=2) {
    angle = 0.5*(float)(i-1)*del;
    trigs[la+i-1] = cos(angle);
    trigs[la+i] = sin(angle);
  }
  if(imode <= 3) 
    return;
  del = 0.5*del;
  la += nn;
  if(*mode == 5) 
    goto S40;
  for(i=2; i<=nn; i++) {
    angle = (float)(i-1)*del;
    trigs[la+i-1] = 2.0*sin(angle);
  }
  return;
S40:
  del = 0.5*del;
  for(i=2; i<=*n; i++) {
    angle = (float)(i-1)*del;
    trigs[la+i-1] = sin(angle);
  }
  return;
}
